import torch
from torch.utils.data import Dataset, TensorDataset


class SubsequenceDataset(Dataset):
    r"""A dataset returning sub-sequences extracted from longer sequences.

    Args:
        *tensors (Tensor): tensors that have the same size of the first dimension.

    Examples:

        >>> u = torch.randn(1000, 2) # 2 inputs
        >>> y = torch.randn(1000, 3) # 3 outputs
        >>> train_dataset = SubsequenceDataset(u, y, subseq_len=100)
    """

    def __init__(self, *tensors, subseq_len):
        self.tensors = tensors
        self.subseq_len = subseq_len
        self.length = self.tensors[0].shape[0]

    def __len__(self):
        return self.length - self.subseq_len

    def __getitem__(self, idx):
        subsequences = [tensor[idx:idx+self.subseq_len] for tensor in self.tensors]
        return subsequences



WH2009 = 'http://www.nonlinearbenchmark.org/FILES/BENCHMARKS/WIENERHAMMERSTEIN/WienerHammerstein2009Files.zip'
CTS = 'http://www.nonlinearbenchmark.org/FILES/BENCHMARKS/CASCADEDTANKS/CascadedTanksFiles.zip'


import torch
import torch.nn as nn
import numpy as np


class NeuralIOSimulator:
    """ This class implements prediction/simulation methods for the IO model structure

     Attributes
     ----------
     io_model: nn.Module
               The neural IO model to be fitted
     """

    def __init__(self, io_model):
        self.io_model = io_model

    def f_onestep(self, PHI):
        """ Naive one-step prediction

        Parameters
        ----------
        PHI : Tensor. Size: (N, n_a + n_b)
              Measured IO regressor tensor

        Returns
        -------
        Tensor. Size: (N, n_y)
            One-step prediction of the output

        """

        Y_pred = self.io_model(PHI)
        return Y_pred

    def f_sim(self, y_seq, u_seq, U):
        """ Open-loop simulation

        Parameters
        ----------
        y_seq: Tensor. Size: (n_a)
               Initial regressor with past values of y

        u_seq: Tensor. Size: (n_b)
               Initial regressor with past values of u

        U : Tensor. Size: (N, n_u)
            Input sequence tensor

        Returns
        -------
        Tensor. Size: (N, n_y)
            Open-loop simulation of the output

        """
        N = np.shape(U)[0]
        Y_list = []

        for i in range(N):
            phi = torch.cat((y_seq, u_seq))
            yi = self.io_model(phi)
            Y_list += [yi]

            if i < N-1:
                # y shift
                y_seq[1:] = y_seq[0:-1]
                y_seq[0] = yi

                # u shift
                u_seq[1:] = u_seq[0:-1]
                u_seq[0] = U[i]

        Y = torch.stack(Y_list, 0)
        return Y

    def f_sim_multistep(self, batch_u, batch_y_seq, batch_u_seq):
        """ Multi-step simulation over (mini)batches

        Parameters
        ----------
        batch_u: Tensor. Size: (q, m, n_u)
                 Input sequence for each subsequence in the minibatch

        batch_y_seq: Tensor. Size: (q, n_a)
                 Initial regressor with past values of y for each subsequence in the minibatch

        batch_u_seq: Tensor. Size: (q, n_b)
                 Initial regressor with past values of u for each subsequence in the minibatch

        Returns
        -------
        Tensor. Size: (q, m, n_y)
            Simulated output for all subsequences in the minibatch

        """

        batch_size = batch_u.shape[0] # number of training samples in the batch
        seq_len = batch_u.shape[1] # length of the training sequences
        n_a = batch_y_seq.shape[1] # number of autoregressive terms on y
        n_b = batch_u_seq.shape[1] # number of autoregressive terms on u

        Y_sim_list = []
        for i in range(seq_len):
            phi = torch.cat((batch_y_seq, batch_u_seq), -1)
            yi = self.io_model(phi)
            Y_sim_list += [yi]

            # y shift
            batch_y_seq[:, 1:] = batch_y_seq[:, 0:-1]
            batch_y_seq[:, [0]] = yi[:]

            # u shift
            batch_u_seq[:, 1:] = batch_u_seq[:, 0:-1]
            batch_u_seq[:, [0]] = batch_u[:, i]

        Y_sim = torch.stack(Y_sim_list, 1)
        return Y_sim



import torch
import torch.nn as nn
from torchid.ss.poly_utils import valid_coeffs


class NeuralStateUpdate(nn.Module):

    r"""State-update mapping modeled as a feed-forward neural network with one hidden layer.

    The model has structure:

    .. math::
        \begin{aligned}
            x_{k+1} = x_k + \mathcal{N}(x_k, u_k),
        \end{aligned}
    where :math:`\mathcal{N}(\cdot, \cdot)` is a feed-forward neural network with one hidden layer.

    Args:
        n_x (int): Number of state variables
        n_u (int): Number of input variables
        n_feat: (int, optional): Number of input features in the hidden layer. Default: 0
        init_small: (boolean, optional): If True, initialize to a Gaussian with mean 0 and std 10^-4. Default: True
        activation: (str): Activation function in the hidden layer. Either 'relu', 'softplus', 'tanh'. Default: 'relu'

    Examples::

        >>> ss_model = NeuralStateUpdate(n_x=2, n_u=1, n_feat=64)
    """

    def __init__(self, n_x, n_u, n_feat=64, init_small=True):
        super(NeuralStateUpdate, self).__init__()
        self.n_x = n_x
        self.n_u = n_u
        self.n_feat = n_feat
        self.net = nn.Sequential(
            nn.Linear(n_x+n_u, n_feat),  # 2 states, 1 input
            nn.ReLU(),
            nn.Linear(n_feat, n_x)
        )

        if init_small:
            for m in self.net.modules():
                if isinstance(m, nn.Linear):
                    nn.init.normal_(m.weight, mean=0, std=1e-4)
                    nn.init.constant_(m.bias, val=0)
    
    def forward(self, X, U):
        XU = torch.cat((X, U), -1)
        DX = self.net(XU)
        return DX


class PolynomialStateUpdate(nn.Module):
    r"""State-update mapping modeled as a polynomial in x and u.

    The model has structure:

    .. math::
        \begin{aligned}
            x_{k+1} = x_k + Ax_{k} + Bu_{k} + Ez_{k},
        \end{aligned}
    where z_{k} is a vector containing (non-linear) monomials in x_{k} and u_{k}

    Args:
        n_x: (np.array): Number of states.
        n_u: (np.array): Number of inputs.
        d_max (int): Maximum degree of the polynomial model.

    """

    def __init__(self, n_x, n_u, d_max, init_small=True):
        super(PolynomialStateUpdate, self).__init__()

        self.n_x = n_x
        self.n_u = n_u
        poly_coeffs = valid_coeffs(n_x + n_u, d_max)
        self.n_poly = len(poly_coeffs)
        self.poly_coeffs = torch.tensor(poly_coeffs)
        self.A = nn.Linear(n_x, n_x, bias=False)
        self.B = nn.Linear(n_u, n_x, bias=False)
        #self.D = nn.linear(n_u, n_y, bias=False)
        self.E = nn.Linear(self.n_poly, n_x, bias=False)
        #self.F = nn.linear(self.n_poly, n_y)

        if init_small:
            nn.init.normal_(self.A.weight, mean=0, std=1e-3)
            nn.init.normal_(self.B.weight, mean=0, std=1e-3)
            nn.init.normal_(self.E.weight, mean=0, std=1e-6)

                # nn.init.constant_(module.bias, val=0)

    def freeze_nl(self):
        self.E.requires_grad_(False)

    def unfreeze_nl(self):
        self.E.requires_grad_(True)

    def forward(self, x, u):
        xu = torch.cat((x, u), dim=-1)
        xu_ = xu.unsqueeze(xu.ndim - 1)
        zeta = torch.prod(torch.pow(xu_, self.poly_coeffs), axis=-1)
        #eta = torch.prod(torch.pow(xu_, self.poly_coeffs), axis=-1)

        dx = self.A(x) + self.B(u) + self.E(zeta)
        return dx


class LinearStateUpdate(nn.Module):
    r"""State-update mapping modeled as a linear function in x and u.

    The model has structure:

    .. math::
        \begin{aligned}
            x_{k+1} = x_k + Ax_{k} + Bu_{k}.
        \end{aligned}

    Args:
        n_x: (np.array): Number of states.
        n_u: (np.array): Number of inputs.
        d_max (int): Maximum degree of the polynomial model.

    """

    def __init__(self, n_x, n_u, init_small=True):
        super(LinearStateUpdate, self).__init__()

        self.n_x = n_x
        self.n_u = n_u
        self.A = nn.Linear(n_x, n_x, bias=False)
        self.B = nn.Linear(n_u, n_x, bias=False)

        if init_small:
            for module in [self.A, self.B]:
                nn.init.normal_(module.weight, mean=0, std=1e-2)
                # nn.init.constant_(module.bias, val=0)

    def forward(self, x, u):
        dx = self.A(x) + self.B(u)
        return dx


class CTSNeuralStateSpace(nn.Module):
    r"""A state-space model to represent the cascaded two-tank system.

    Args:
        n_feat: (int, optional): Number of input features in the hidden layer. Default: 0
        init_small: (boolean, optional): If True, initialize to a Gaussian with mean 0 and std 10^-4. Default: True
    """

    def __init__(self, n_x, n_u, n_feat=64, init_small=True):
        super(CTSNeuralStateSpace, self).__init__()
        self.n_x = n_x
        self.n_u = n_u
        self.n_feat = n_feat
        self.net = nn.Sequential(
            nn.Linear(n_x + n_u, n_feat),  # 2 states, 1 input
            nn.ReLU(),
            nn.Linear(n_feat, n_x)
        )

        if init_small:
            for m in self.net.modules():
                if isinstance(m, nn.Linear):
                    nn.init.normal_(m.weight, mean=0, std=1e-4)
                    nn.init.constant_(m.bias, val=0)

    def forward(self, X, U):
        XU = torch.cat((X, U), -1)
        DX = self.net(XU)
        return DX


class LinearOutput(nn.Module):
    r"""Output  mapping modeled as a linear function in x.

    The model has structure:

    .. math::
        \begin{aligned}
            y_{k} = Cx_k.
        \end{aligned}
    """

    def __init__(self, n_x, n_y, bias=False):
        super(LinearOutput, self).__init__()
        self.n_x = n_x
        self.n_y = n_y
        self.C = torch.nn.Linear(n_x, n_y, bias=bias)

    def forward(self, x):
        return self.C(x)


class ChannelsOutput(nn.Module):
    r"""Output  mapping corresponding to a specific state channel.

    """
    def __init__(self, channels):
        super(ChannelsOutput, self).__init__()
        self.channels = channels

    def forward(self, x):
        y = x[..., self.channels]
        return y



import torch
import torch.nn as nn
import numpy as np
import nodepy
from typing import List


class ForwardEulerSimulator(nn.Module):

    r""" Forward Euler integration of a continuous-time neural state space model.

    Args:
        ss_model (nn.Module): The neural state-space model.
        ts (np.float): Sampling time for simulation.
        batch_first (bool): If True, first dimension is batch.

    Inputs: x_0, input
        * **x_0**: tensor of shape :math:`(N, n_{x})` containing the
          initial hidden state for each element in the batch.
          Defaults to zeros if (h_0, c_0) is not provided.
        * **input**: tensor of shape :math:`(L, N, n_{u})` when ``batch_first=False`` or
          :math:`(N, L, n_{x})` when ``batch_first=True`` containing the input sequence

    Outputs: states
        * **states**: tensor of shape :math:`(L, N, n_{x})` corresponding to
          the simulated state sequence.

    Examples::

        >>> ss_model = NeuralStateSpaceModel(n_x=3, n_u=2)
        >>> nn_solution = ForwardEulerSimulator(ss_model)
        >>> x0 = torch.randn(64, 3)
        >>> u = torch.randn(100, 64, 2)
        >>> x = nn_solution(x0, u)
        >>> print(x.size())
        torch.Size([100, 64, 3])
     """

    def __init__(self, ss_model, ts=1.0, batch_first=False):
        super(ForwardEulerSimulator, self).__init__()
        self.ss_model = ss_model
        self.ts = ts
        self.batch_first = batch_first

    def forward(self, x_0: torch.Tensor, input: torch.Tensor) -> torch.Tensor:

        states: List[torch.Tensor] = []
        x_step = x_0
        dim_time = 1 if self.batch_first else 0

        for u_step in input.split(1, dim=dim_time):  # split along the time axis
            u_step = u_step.squeeze(dim_time)
            states += [x_step]
            dx = self.ss_model(x_step, u_step)
            x_step = x_step + self.ts*dx

        states = torch.stack(states, dim_time)
        return states


class ExplicitRKSimulator(nn.Module):
    """ This class implements prediction/simulation methods for a continuous SS model structure

     Attributes
     ----------
     ss_model: nn.Module
               The neural SS model to be fitted
     ts: float
         model sampling time (when it is fixed)

     scheme: string
          Runge-Kutta scheme to be used
    """

    def __init__(self, ss_model, ts=1.0, scheme='RK44'):
        super(ExplicitRKSimulator, self).__init__()
        self.ss_model = ss_model
        self.ts = ts
        info_RK = nodepy.runge_kutta_method.loadRKM(scheme)
        self.A = torch.FloatTensor(info_RK.A.astype(np.float32))
        self.b = torch.FloatTensor(info_RK.b.astype(np.float32))
        self.c = torch.FloatTensor(info_RK.c.astype(np.float32))
        self.stages = self.b.numel()  # number of stages of the rk method

    def forward(self, x0_batch, u_batch):
        """ Multi-step simulation over (mini)batches

        Parameters
        ----------
        x0_batch: Tensor. Size: (q, n_x)
             Initial state for each subsequence in the minibatch

        u_batch: Tensor. Size: (m, q, n_u)
            Input sequence for each subsequence in the minibatch

        Returns
        -------
        Tensor. Size: (m, q, n_x)
            Simulated state for all subsequences in the minibatch

        """

        batch_size = x0_batch.shape[0]
        n_x = x0_batch.shape[1]
        seq_len = u_batch.shape[0]

        X_sim_list = []
        x_step = x0_batch
        for u_step in u_batch.split(1):#i in range(seq_len):

            u_step = u_step.squeeze(0)
            X_sim_list += [x_step]
            #u_step = u_batch[i, :, :]

            K = []  #torch.zeros((self.stages, nx))
            for stage_idx in range(self.stages):  # compute Ki, i=0,1,..s-1
                DX_pred = torch.zeros((batch_size, n_x)).to(self.device)
                for j in range(stage_idx):  # j=0,1,...i-1
                    DX_pred = DX_pred +  self.A[stage_idx, j] * K[j]
                DX_pred = DX_pred*self.ts
                K.append(self.ss_model(x_step + DX_pred, u_step))  # should u be interpolated??
            F = torch.zeros((batch_size, n_x)).to(self.device)
            for stage_idx in range(self.stages):
                F += self.b[stage_idx]*K[stage_idx]
            x_step = x_step + self.ts*F

        X_sim = torch.stack(X_sim_list, 0)

        return X_sim


class RK4Simulator(nn.Module):
    """ This class implements prediction/simulation methods for a continuous SS model structure

     Attributes
     ----------
     ss_model: nn.Module
               The neural SS model to be fitted
     ts: float
         model sampling time (when it is fixed)

     scheme: string
          Runge-Kutta scheme to be used
    """

    def __init__(self, ss_model, ts=1.0):
        super(RK4Simulator, self).__init__()
        self.ss_model = ss_model
        self.ts = ts

    def forward(self, x0_batch, u_batch):
        """ Multi-step simulation over (mini)batches

        Parameters
        ----------
        x0_batch: Tensor. Size: (q, n_x)
             Initial state for each subsequence in the minibatch

        u_batch: Tensor. Size: (m, q, n_u)
            Input sequence for each subsequence in the minibatch

        Returns
        -------
        Tensor. Size: (m, q, n_x)
            Simulated state for all subsequences in the minibatch

        """

        X_sim_list = []
        x_step = x0_batch
        for u_step in u_batch.split(1):

            u_step = u_step.squeeze(0)
            X_sim_list += [x_step]

            dt2 = self.ts / 2.0
            k1 = self.ss_model(x_step, u_step)
            k2 = self.ss_model(x_step + dt2 * k1, u_step)
            k3 = self.ss_model(x_step + dt2 * k2, u_step)
            k4 = self.ss_model(x_step + self.ts * k3, u_step)
            dx = self.ts / 6.0 * (k1 + 2.0 * k2 + 2.0 * k3 + k4)
            x_step = x_step + dx

        X_sim = torch.stack(X_sim_list, 0)

        return X_sim


import torch
import torch.nn as nn
from typing import List

        
class StateSpaceSimulator(nn.Module):
    r""" Discrete-time state-space simulator.

    Args:
        f_xu (nn.Module): The neural state-space model.
        batch_first (bool): If True, first dimension is batch.

    Inputs: x_0, u
        * **x_0**: tensor of shape :math:`(N, n_{x})` containing the
          initial hidden state for each element in the batch.
          Defaults to zeros if (h_0, c_0) is not provided.
        * **input**: tensor of shape :math:`(L, N, n_{u})` when ``batch_first=False`` or
          :math:`(N, L, n_{x})` when ``batch_first=True`` containing the input sequence

    Outputs: x
        * **x**: tensor of shape :math:`(L, N, n_{x})` corresponding to
          the simulated state sequence.

    Examples::

        >>> ss_model = NeuralStateSpaceModel(n_x=3, n_u=2)
        >>> nn_solution = StateSpaceSimulator(ss_model)
        >>> x0 = torch.randn(64, 3)
        >>> u = torch.randn(100, 64, 2)
        >>> x = nn_solution(x0, u)
        >>> print(x.size())
        torch.Size([100, 64, 3])
     """

    def __init__(self, f_xu, g_x=None, batch_first=False):
        super().__init__()
        self.state_update = f_xu
        self.output = g_x
        self.batch_first = batch_first

    def simulate_state(self, x_0, u):
        x: List[torch.Tensor] = []
        x_step = x_0
        dim_time = 1 if self.batch_first else 0

        for u_step in u.split(1, dim=dim_time):  # split along the time axis
            u_step = u_step.squeeze(dim_time)
            x += [x_step]
            dx = self.state_update(x_step, u_step)
            x_step = x_step + dx

        x = torch.stack(x, dim_time)
        return x

    def forward(self, x_0, u, return_x=False):
        x = self.simulate_state(x_0, u)
        if self.output is not None:
            y = self.output(x)
        else:
            y = x
        if not return_x:
            return y
        else:
            return y, x





import torch
import torch.nn as nn


class FlippedLSTMStateEstimator(nn.Module):
    """ Black-box estimator from the sequences of (u, y) to x[0].
    The estimation is performed by processing (u, y) backward in time.
    """

    def __init__(self, n_u=1, n_y=1, n_x=2, hidden_size=16, batch_first=False):
        super(FlippedLSTMStateEstimator, self).__init__()
        self.n_u = n_u
        self.n_y = n_y
        self.n_x = n_x
        self.batch_first = batch_first

        self.lstm = nn.LSTM(input_size=n_y + n_u, hidden_size=hidden_size, batch_first=batch_first)
        self.lstm_output = nn.Linear(hidden_size, n_x)

    def forward(self, u, y):
        uy = torch.cat((u, y), -1)
        uy_rev = uy.flip(0)
        h_rev, (h0, c0) = self.lstm(uy_rev)
        x0 = self.lstm_output(h0).squeeze(0)
        return x0



import torch
from torchid.ss.dt.models import NeuralStateUpdate
from torchid.ss.dt.simulator import StateSpaceSimulator


def test_batchfirst():
    """Test batch-first option of ForwardEulerSimulator"""

    L = 100  # sequence length
    N = 64  # batch size
    n_x = 2  # states
    n_u = 3

    f_xu = NeuralStateUpdate(n_x=n_x, n_u=n_u)
    model_tf = StateSpaceSimulator(f_xu)  # batch_first=False, thus time first, default
    model_bf = StateSpaceSimulator(f_xu, batch_first=True)

    x0 = torch.randn(N, n_x)
    u_tf = torch.randn(L, N, n_u)
    u_bf = torch.transpose(u_tf, 0, 1)  # transpose time/batch dimensions

    x_tf = model_tf(x0, u_tf)
    x_bf = model_bf(x0, u_bf)

    assert(torch.allclose(x_bf.transpose(0, 1), x_tf))

